@@ -1,6 +1,5 @@
 /*
- * Copyright (C) 2011-2016 Canonical
- * Hugely modified parts from powertop-1.13, Copyright 2007, Intel Corporation
+ * Copyright (C) 2011-2017 Canonical
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -40,6 +39,7 @@
 #include <math.h>
 #include <float.h>
 #include <ncurses.h>
+#include <ctype.h>
 
 #define TABLE_SIZE		(1009)		/* Should be a prime */
 
@@ -55,13 +55,17 @@
 #define OPT_USER		(0x00000200)
 #define OPT_SHOW_WHENCE		(0x00000400)
 #define OPT_TOP			(0x00000800)
+#define OPT_TIMER_ID		(0x00001000)
 #define OPT_CMD			(OPT_CMD_SHORT | OPT_CMD_LONG)
 
-#define EVENT_BUF_SIZE		(8192)
+#define EVENT_BUF_SIZE		(64 * 1024)
 #define TIMER_REAP_AGE		(600)	/* Age of timer before it is reaped */
 #define TIMER_REAP_THRESHOLD	(30)
+#define EVENTS_WIDTH		(8)
 #define TASK_WIDTH		(15)
-#define FUNC_WIDTH		(25)
+#define TIMER_ID_WIDTH		(16)
+#define FUNC_WIDTH		(24)
+#define FUNC_WIDTH_MAX		(30)
 
 #define _VER_(major, minor, patchlevel) \
 	((major * 10000) + (minor * 100) + patchlevel)
@@ -94,6 +98,9 @@
 #define FLOAT_TINY		(0.0000001)
 #define FLOAT_CMP(a, b)		(fabs(a - b) < FLOAT_TINY)
 
+/*
+ *  timer_info_t contains per task timer infos.
+ */
 typedef struct timer_info {
 	struct timer_info *next;	/* Next in list */
 	struct timer_info *hash_next;	/* Next in hash list */
@@ -102,30 +109,30 @@ typedef struct timer_info {
 	char 		*task_mangled;	/* Modified name of process/kernel */
 	char		*cmdline;	/* From /proc/$pid/cmdline */
 	char		*func;		/* Kernel waiting func */
-	char		*callback;	/* Kernel timer callback func */
 	char		*ident;		/* Unique identity */
 	bool		kernel_thread;	/* True if task is a kernel thread */
-	uint32_t	ref_count;	/* timer stat reference count */
-	uint64_t	total;		/* Total number of events */
+	uint32_t	ref_count;	/* Timer stat reference count */
+	uint64_t	timer;		/* Timer ID */
+	uint64_t	total_events;	/* Total number of events */
+	uint64_t	delta_events;	/* Events in one time period */
 	double		time_total;	/* Total time */
 	double		last_used;	/* Last referenced */
+	double		prev_used;	/* Previous time used */
 } timer_info_t;
 
 typedef struct timer_stat {
 	struct timer_stat *next;	/* Next timer stat in hash table */
-	struct timer_stat *sorted_freq_next; /* Next timer stat in event frequency sorted list */
-	uint64_t	count;		/* Number of events */
-	int64_t		delta;		/* Change in events since last time */
-	double		time;		/* Time of sample */
-	double		time_delta;	/* Change in time since last time */
+	struct timer_stat *sorted_freq_next; /* Next timer stat in event */
+					/* frequency sorted list */
 	timer_info_t	*info;		/* Timer info */
 } timer_stat_t;
 
 /* sample delta item as an element of the sample_delta_list_t */
 typedef struct sample_delta_item {
 	struct sample_delta_item *next;	/* next in list */
-	int64_t		delta;		/* difference in timer events between old and new */
-	double		time_delta;	/* difference in time between old and new */
+	int64_t		delta_events;	/* delta in events */
+	double		time_delta;	/* difference in time between old */
+					/* and new */
 	timer_info_t	*info;		/* timer this refers to */
 } sample_delta_item_t;
 
@@ -143,18 +150,33 @@ typedef struct {
 
 #define KERN_TASK_INFO(str)		{ str, sizeof(str) - 1 }
 
-static const char *app_name = "eventstat";
-static const char *proc_timer_stats = "/proc/timer_stats";
+static const char * const app_name = "eventstat";
+static const char * const sys_tracing_enable =
+	"/sys/kernel/debug/tracing/events/timer/hrtimer_start/enable";
+static const char * const sys_tracing_pipe =
+	"/sys/kernel/debug/tracing/trace_pipe";
+static const char * const sys_tracing_set_event =
+	"/sys/kernel/debug/tracing/set_event";
+static const char * const sys_tracing_filter =
+	"/sys/kernel/debug/tracing/events/timer/filter";
 
 static void es_printf(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
 
 static timer_stat_t *timer_stat_free_list; /* free list of timer stats */
 static timer_info_t *timer_info_list;	/* cache list of timer_info */
 static timer_info_t *timer_info_hash[TABLE_SIZE]; /* hash of timer_info */
-static sample_delta_list_t *sample_delta_list;	/* list of samples, sorted in sample time order */
+
+/* head of list of samples, sorted in sample time order */
+static sample_delta_list_t *sample_delta_list_head;
+
+/* tail of list of samples, sorted in sample time order */
+static sample_delta_list_t *sample_delta_list_tail;
+
+/* ignore samples with event delta less than this */
+static double opt_threshold;
+
 static char *csv_results;		/* results in comma separated values */
 static char *get_events_buf;		/* buffer to glob events into */
-static double  opt_threshold;		/* ignore samples with event delta less than this */
 static uint32_t timer_info_list_length;	/* length of timer_info_list */
 static uint32_t opt_flags;		/* option flags */
 static volatile bool stop_eventstat = false;	/* set by sighandler */
@@ -217,6 +239,42 @@ static const int signals[] = {
 	-1,
 };
 
+/*
+ *  pid_max_digits()
+ *	determine (or guess) maximum digits of pids
+ */
+static int pid_max_digits(void)
+{
+	static int max_digits;
+	ssize_t n;
+	int fd;
+	const int default_digits = 6;
+	const int min_digits = 5;
+	char buf[32];
+
+	if (max_digits)
+		goto ret;
+
+	max_digits = default_digits;
+	fd = open("/proc/sys/kernel/pid_max", O_RDONLY);
+	if (fd < 0)
+		goto ret;
+	n = read(fd, buf, sizeof(buf) - 1);
+	(void)close(fd);
+	if (n < 0)
+		goto ret;
+
+	buf[n] = '\0';
+	max_digits = 0;
+	while ((buf[max_digits] >= '0') && (buf[max_digits] <= '9'))
+		max_digits++;
+	if (max_digits < min_digits)
+		max_digits = min_digits;
+ret:
+	return max_digits;
+
+}
+
 /*
  *  hash_djb2a()
  *	Hash a string, from Dan Bernstein comp.lang.c (xor version)
@@ -229,6 +287,7 @@ static HOT OPTIMIZE3 uint32_t hash_djb2a(const char *str)
 	while ((c = *str++)) {
 		/* (hash * 33) ^ c */
 		hash = ((hash << 5) + hash) ^ c;
+		hash = (hash * 33) ^ c;
 	}
 	return hash % TABLE_SIZE;
 }
@@ -237,7 +296,7 @@ static HOT OPTIMIZE3 uint32_t hash_djb2a(const char *str)
  *  eventstat_winsize()
  *	get tty size
  */
-static void eventstat_winsize()
+static void eventstat_winsize(void)
 {
 	struct winsize ws;
 
@@ -247,10 +306,9 @@ static void eventstat_winsize()
 	}
 }
 
-
 /*
  *  eventstat_clear()
- *  	clear screen if in top mode
+ *	clear screen if in top mode
  */
 static inline void eventstat_clear(void)
 {
@@ -262,7 +320,7 @@ static inline void eventstat_clear(void)
  *  eventstat_refresh()
  *	refresh screen if in top mode
  */
-static inline void eventstat_refresh()
+static inline void eventstat_refresh(void)
 {
 	if (curses_init)
 		refresh();
@@ -270,11 +328,9 @@ static inline void eventstat_refresh()
 
 /*
  *  eventstat_move()
- * 	move cursor if in top mode
+ *	move cursor if in top mode
  */
-static inline void eventstat_move(
-	const int y,
-	const int x)
+static inline void eventstat_move(const int y, const int x)
 {
 	if (curses_init)
 		move(y, x);
@@ -296,7 +352,8 @@ static void eventstat_endwin(void)
  *  err_abort()
  *	print an error and exit
  */
-static void __attribute__ ((noreturn)) err_abort(const char *fmt, ...)
+static void __attribute__ ((noreturn)) __attribute__((format(printf, 1, 2)))
+err_abort(const char *fmt, ...)
 {
 	va_list ap;
 
@@ -309,39 +366,63 @@ static void __attribute__ ((noreturn)) err_abort(const char *fmt, ...)
 }
 
 /*
- *  set_timer_stat()
+ *  set_tracing_enable()
  *	enable/disable timer stat
  */
-static void set_timer_stat(const char *str, const bool carp)
+static void set_tracing(const char *path, const char *str, const bool carp)
 {
 	int fd;
 	const ssize_t len = (ssize_t)strlen(str);
 
-	if ((fd = open(proc_timer_stats, O_WRONLY, S_IRUSR | S_IWUSR)) < 0) {
+	if ((fd = open(path, O_WRONLY, S_IRUSR | S_IWUSR)) < 0) {
 		if (carp) {
 			err_abort("Cannot open %s, errno=%d (%s)\n",
-				proc_timer_stats, errno, strerror(errno));
+				path, errno, strerror(errno));
 		}
 		return;
 	}
 	if (write(fd, str, len) != len) {
-		close(fd);
+		(void)close(fd);
 		if (carp) {
 			err_abort("Cannot write to %s, errno=%d (%s)\n",
-				proc_timer_stats, errno, strerror(errno));
+				path, errno, strerror(errno));
 		}
 		return;
 	}
 	(void)close(fd);
 }
 
+/*
+ *  set_tracing_enable()
+ *	enable/disable timer stat
+ */
+static inline void set_tracing_enable(const char *str, const bool carp)
+{
+	set_tracing(sys_tracing_enable, str, carp);
+}
+
+static void set_tracing_event(void)
+{
+	char buffer[64];
+
+	set_tracing(sys_tracing_set_event, "\n", true);
+	set_tracing(sys_tracing_set_event, "hrtimer_start", true);
+	set_tracing(sys_tracing_filter, "0", true);
+
+	/* Ignore event stat and idle events */
+	snprintf(buffer, sizeof(buffer),
+		"common_pid != %d && common_pid != 0", getpid());
+	set_tracing(sys_tracing_filter, buffer, true);
+}
+
+
 /*
  *  eventstat_exit()
  *	exit and set timer stat to 0
  */
 static void __attribute__ ((noreturn)) eventstat_exit(const int status)
 {
-	set_timer_stat("0\n", false);
+	set_tracing_enable("0\n", false);
 	exit(status);
 }
 
@@ -432,7 +513,7 @@ static void handle_sig(int dummy)
 	(void)dummy;	/* Stop unused parameter warning with -Wextra */
 
 	stop_eventstat = true;
-	set_timer_stat("0\n", false);
+	set_tracing_enable("0\n", false);
 }
 
 /*
@@ -441,7 +522,7 @@ static void handle_sig(int dummy)
  */
 static inline void samples_free(void)
 {
-	sample_delta_list_t *sdl = sample_delta_list;
+	sample_delta_list_t *sdl = sample_delta_list_head;
 
 	while (sdl) {
 		sample_delta_list_t *sdl_next = sdl->next;
@@ -461,18 +542,16 @@ static inline void samples_free(void)
  *	add a timer_stat's delta and info field to a
  *	list at time position whence
  */
-static void sample_add(
-	timer_stat_t *timer_stat,
-	const double whence)
+static void sample_add(timer_stat_t *timer_stat, const double whence)
 {
 	bool	found = false;
 	sample_delta_list_t *sdl;
 	sample_delta_item_t *sdi;
 
-	if (csv_results == NULL)	/* No need if not request */
+	if (csv_results == NULL)	/* No need if not enabled */
 		return;
 
-	for (sdl = sample_delta_list; sdl; sdl = sdl->next) {
+	for (sdl = sample_delta_list_head; sdl; sdl = sdl->next) {
 		if (FLOAT_CMP(sdl->whence, whence)) {
 			found = true;
 			break;
@@ -487,15 +566,22 @@ static void sample_add(
 		if ((sdl = calloc(1, sizeof(sample_delta_list_t))) == NULL)
 			err_abort("Cannot allocate sample delta list\n");
 		sdl->whence = whence;
-		sdl->next = sample_delta_list;
-		sample_delta_list = sdl;
+
+		if (sample_delta_list_tail) {
+			sample_delta_list_tail->next = sdl;
+			sample_delta_list_tail = sdl;
+		} else {
+			sample_delta_list_head = sdl;
+			sample_delta_list_tail = sdl;
+		}
 	}
 
 	/* Now append the sdi onto the list */
 	if ((sdi = calloc(1, sizeof(sample_delta_item_t))) == NULL)
 		err_abort("Cannot allocate sample delta item\n");
-	sdi->delta = timer_stat->delta;
-	sdi->time_delta = timer_stat->time_delta;
+	sdi->delta_events = timer_stat->info->delta_events;
+	sdi->time_delta = timer_stat->info->last_used -
+			  timer_stat->info->prev_used;
 	sdi->info = timer_stat->info;
 	sdi->next = sdl->list;
 	sdl->list = sdi;
@@ -528,10 +614,10 @@ static int info_compare_total(const void *item1, const void *item2)
 	timer_info_t *const *info1 = (timer_info_t *const *)item1;
 	timer_info_t *const *info2 = (timer_info_t *const *)item2;
 
-	if ((*info2)->total == (*info1)->total)
+	if ((*info2)->total_events == (*info1)->total_events)
 		return 0;
 
-	return ((*info2)->total > (*info1)->total) ? 1 : -1;
+	return ((*info2)->total_events > (*info1)->total_events) ? 1 : -1;
 }
 
 static bool pid_a_kernel_thread_guess(const char *task)
@@ -567,6 +653,7 @@ static bool pid_a_kernel_thread_guess(const char *task)
 		KERN_TASK_INFO("bdi-default-"),
 		{ NULL, 0 }
 	};
+
 	size_t i;
 
 	for (i = 0; kernel_tasks[i].task != NULL; i++) {
@@ -580,19 +667,12 @@ static bool pid_a_kernel_thread_guess(const char *task)
  *  pid_a_kernel_thread
  *
  */
-static bool pid_a_kernel_thread(
-	const char *cmdline,
-	const char *task,
-	const pid_t id)
+static bool pid_a_kernel_thread(const char *task, const pid_t id)
 {
-	pid_t pgid;
-
-	if (sane_procs && (cmdline != NULL))
-		return (*cmdline == '\0');
+	const pid_t pgid = id == 0 ? 0 : getpgid(id);
 
 	/* We are either in a container, or with a task with a NULL cmdline */
-	pgid = getpgid(id);
-	if (pgid >= 0)
+	if (sane_procs || (id >= 0))
 		return (pgid == 0);
 
 	/* Can't get pgid on that pid, so make a guess */
@@ -601,7 +681,7 @@ static bool pid_a_kernel_thread(
 
 /*
  *  get_pid_cmdline
- * 	get process's /proc/pid/cmdline
+ *	get process's /proc/pid/cmdline
  */
 static char *get_pid_cmdline(const pid_t id)
 {
@@ -662,8 +742,7 @@ static inline double duration_round(const double duration)
 static void samples_dump(const char *filename)
 {
 	timer_info_t **sorted_timer_infos;
-	size_t i = 0;
-	size_t n;
+	size_t i, n;
 	FILE *fp;
 	uint64_t count = 0;
 	double first_time = -1.0;
@@ -678,16 +757,19 @@ static void samples_dump(const char *filename)
 		return;
 	}
 
-	if ((sorted_timer_infos = calloc(timer_info_list_length, sizeof(timer_info_t*))) == NULL)
+	sorted_timer_infos = calloc(timer_info_list_length,
+				sizeof(timer_info_t *));
+	if (!sorted_timer_infos)
 		err_abort("Cannot allocate buffer for sorting timer_infos\n");
 
 	/* Just want the timers with some non-zero total */
 	for (n = 0, info = timer_info_list; info; info = info->next) {
-		if (info->total > 0)
+		if (info->total_events > 0)
 			sorted_timer_infos[n++] = info;
 	}
 
-	qsort(sorted_timer_infos, n, sizeof(timer_info_t *), info_compare_total);
+	qsort(sorted_timer_infos, n,
+		sizeof(timer_info_t *), info_compare_total);
 
 	fprintf(fp, "Time:,Task:");
 	for (i = 0; i < n; i++) {
@@ -707,17 +789,12 @@ static void samples_dump(const char *filename)
 		fprintf(fp, ",%s", sorted_timer_infos[i]->func);
 	fprintf(fp, "\n");
 
-	fprintf(fp, ",Callback:");
-	for (i = 0; i < n; i++)
-		fprintf(fp, ",%s", sorted_timer_infos[i]->callback);
-	fprintf(fp, "\n");
-
 	fprintf(fp, ",Total:");
 	for (i = 0; i < n; i++)
-		fprintf(fp, ",%" PRIu64, sorted_timer_infos[i]->total);
+		fprintf(fp, ",%" PRIu64, sorted_timer_infos[i]->total_events);
 	fprintf(fp, "\n");
 
-	for (sdl = sample_delta_list; sdl; sdl= sdl->next) {
+	for (sdl = sample_delta_list_head; sdl; sdl = sdl->next) {
 		time_t t = (time_t)sdl->whence;
 		struct tm *tm;
 
@@ -730,7 +807,9 @@ static void samples_dump(const char *filename)
 			first_time = sdl->whence;
 		fprintf(fp, ",%f", duration_round(sdl->whence - first_time));
 
-		/* Scan in timer info order to be consistent for all sdl rows */
+		/*
+		 * Scan in timer info order to be consistent for all sdl rows
+		 */
 		for (i = 0; i < n; i++) {
 			sample_delta_item_t *sdi =
 				sample_find(sdl, sorted_timer_infos[i]);
@@ -741,8 +820,9 @@ static void samples_dump(const char *filename)
 			 */
 			if (sdi) {
 				double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
-				fprintf(fp, ",%f", FLOAT_CMP(duration, 0.0) ? 0.0 :
-					(double)sdi->delta / duration);
+				fprintf(fp, ",%f",
+					FLOAT_CMP(duration, 0.0) ? -99.99 :
+					(double)sdi->delta_events / duration);
 			} else
 				fprintf(fp, ",%f", 0.0);
 		}
@@ -757,14 +837,14 @@ static void samples_dump(const char *filename)
 		for (i = 0; i < n; i++) {
 			double min = DBL_MAX;
 
-			for (sdl = sample_delta_list; sdl; sdl = sdl->next) {
+			for (sdl = sample_delta_list_head; sdl; sdl = sdl->next) {
 				sample_delta_item_t *sdi =
 					sample_find(sdl, sorted_timer_infos[i]);
 
 				if (sdi) {
 					double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
 					double val = FLOAT_CMP(duration, 0.0) ?
-						0.0 : sdi->delta / duration;
+						0.0 : sdi->delta_events / duration;
 					if (min > val)
 						min = val;
 				}
@@ -777,14 +857,14 @@ static void samples_dump(const char *filename)
 		for (i = 0; i < n; i++) {
 			double max = DBL_MIN;
 
-			for (sdl = sample_delta_list; sdl; sdl= sdl->next) {
+			for (sdl = sample_delta_list_head; sdl; sdl = sdl->next) {
 				sample_delta_item_t *sdi =
 					sample_find(sdl, sorted_timer_infos[i]);
 
 				if (sdi) {
 					double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
 					double val = FLOAT_CMP(duration, 0.0) ?
-						0.0 : sdi->delta / duration;
+						0.0 : sdi->delta_events / duration;
 					if (max < val)
 						max = val;
 				}
@@ -796,7 +876,7 @@ static void samples_dump(const char *filename)
 		fprintf(fp, ",Average:");
 		for (i = 0; i < n; i++)
 			fprintf(fp, ",%f", count == 0 ? 0.0 :
-				(double)sorted_timer_infos[i]->total / count);
+				(double)sorted_timer_infos[i]->total_events / count);
 		fprintf(fp, "\n");
 
 		/*
@@ -805,16 +885,16 @@ static void samples_dump(const char *filename)
 		fprintf(fp, ",Std.Dev.:");
 		for (i = 0; i < n; i++) {
 			double average = (double)
-				sorted_timer_infos[i]->total / (double)count;
+				sorted_timer_infos[i]->total_events / (double)count;
 			double sum = 0.0;
 
-			for (sdl = sample_delta_list; sdl; sdl = sdl->next) {
+			for (sdl = sample_delta_list_head; sdl; sdl = sdl->next) {
 				sample_delta_item_t *sdi =
 					sample_find(sdl, sorted_timer_infos[i]);
 				if (sdi) {
 					double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
 					double diff = FLOAT_CMP(duration, 0.0) ? 0.0 :
-						((double)sdi->delta - average) / duration;
+						((double)sdi->delta_events - average) / duration;
 					diff = diff * diff;
 					sum += diff;
 				}
@@ -824,7 +904,6 @@ static void samples_dump(const char *filename)
 		}
 		fprintf(fp, "\n");
 	}
-
 	free(sorted_timer_infos);
 	(void)fclose(fp);
 }
@@ -837,18 +916,21 @@ static void samples_dump(const char *filename)
 static HOT timer_info_t *timer_info_find(
 	const timer_info_t *new_info,
 	const char *ident,
-	const double time_now)
+	const double time_now,
+	const double duration)
 {
 	timer_info_t *info;
 	const uint32_t h = hash_djb2a(ident);
 
 	for (info = timer_info_hash[h]; info; info = info->hash_next) {
 		if (strcmp(ident, info->ident) == 0) {
+			info->prev_used = info->last_used;
 			info->last_used = time_now;
 			return info;
 		}
 	}
-	if ((info = calloc(1, sizeof(timer_info_t))) == NULL)
+	info = calloc(1, sizeof(timer_info_t));
+	if (!info)
 		err_abort("Cannot allocate timer info\n");
 
 	info->pid = new_info->pid;
@@ -856,19 +938,22 @@ static HOT timer_info_t *timer_info_find(
 	info->task_mangled = strdup(new_info->task_mangled);
 	info->cmdline = strdup(new_info->cmdline);
 	info->func = strdup(new_info->func);
-	info->callback = strdup(new_info->callback);
 	info->ident = strdup(ident);
 	info->kernel_thread = new_info->kernel_thread;
-	info->total = new_info->total;
+	info->total_events = new_info->total_events;
+	info->delta_events = new_info->delta_events;
 	info->time_total = new_info->time_total;
-	info->last_used = time_now;
+	info->timer = new_info->timer;
 	info->ref_count = 0;
+	info->prev_used = time_now - duration;		/* Fake previous time */
+	info->last_used = time_now;
+	info->total_events = 1;
+	info->delta_events = 1;
 
 	if (info->task == NULL ||
 	    info->task_mangled == NULL ||
 	    info->cmdline == NULL ||
 	    info->func == NULL ||
-	    info->callback == NULL ||
 	    info->ident == NULL) {
 		err_abort("Out of memory allocating a timer stat fields\n");
 	}
@@ -896,7 +981,6 @@ static void timer_info_free(void *data)
 		free(info->cmdline);
 	free(info->task_mangled);
 	free(info->func);
-	free(info->callback);
 	free(info->ident);
 	free(info);
 }
@@ -967,7 +1051,7 @@ static void timer_info_purge_old_from_hash_list(
 
 /*
  *  timer_info_purge_old()
- *  	clean out old timer infos
+ *	clean out old timer infos
  */
 static inline void timer_info_purge_old(const double time_now)
 {
@@ -1006,11 +1090,15 @@ static inline void timer_info_list_free(void)
  */
 static char *make_hash_ident(const timer_info_t *info)
 {
-	static char ident[1024];
+	static char ident[128];
 
-	snprintf(ident, sizeof(ident), "%x%s%8.8s%8.8s%s",
-		info->pid, info->task, info->func,
-		info->callback, info->cmdline);
+	if (opt_flags & OPT_TIMER_ID) {
+		snprintf(ident, sizeof(ident), "%x%s%8.8s%" PRIx64,
+			info->pid, info->task, info->func, info->timer);
+	} else {
+		snprintf(ident, sizeof(ident), "%x%s%8.8s",
+			info->pid, info->task, info->func);
+	}
 	return ident;
 }
 
@@ -1035,10 +1123,9 @@ static void timer_stat_free_list_free(void)
  *  timer_stat_free_contents()
  *	Free timers from a hash table
  */
-static void timer_stat_free_contents(
-	timer_stat_t *timer_stats[])	/* timer stat hash table */
+static void timer_stat_free_contents(timer_stat_t *timer_stats[])
 {
-	int i;
+	size_t i;
 
 	for (i = 0; i < TABLE_SIZE; i++) {
 		timer_stat_t *ts = timer_stats[i];
@@ -1065,8 +1152,9 @@ static void timer_stat_free_contents(
  */
 static void timer_stat_add(
 	timer_stat_t *timer_stats[],	/* timer stat hash table */
+	timer_info_t *info,		/* timer info to be added */
 	const double time_now,		/* time sample was taken */
-	timer_info_t *info)		/* timer info to be added */
+	const double duration)		/* duration of a sample */
 {
 	const char *ident = make_hash_ident(info);
 	const uint32_t h = hash_djb2a(ident);
@@ -1074,7 +1162,9 @@ static void timer_stat_add(
 
 	for (ts = timer_stats[h]; ts; ts = ts->next) {
 		if (strcmp(ts->info->ident, ident) == 0) {
-			ts->count += info->total;
+			ts->info->total_events++;
+			ts->info->delta_events++;
+			sample_add(ts, time_now);
 			return;
 		}
 	}
@@ -1089,32 +1179,13 @@ static void timer_stat_add(
 			err_abort("Out of memory allocating a timer stat\n");
 	}
 
-	ts_new->count = info->total;
-	ts_new->info = timer_info_find(info, ident, time_now);
+	ts_new->info = timer_info_find(info, ident, time_now, duration);
 	ts_new->info->ref_count++;
 	ts_new->next = timer_stats[h];
-	ts_new->time = time_now;
 	ts_new->sorted_freq_next = NULL;
+
 	timer_stats[h] = ts_new;
-}
-
-/*
- *  timer_stat_find()
- *	find a timer stat (needle) in a timer stat hash table (haystack)
- */
-static OPTIMIZE3 timer_stat_t *timer_stat_find(
-	timer_stat_t *haystack[],	/* timer stat hash table */
-	timer_stat_t *needle)		/* timer stat to find */
-{
-	timer_stat_t *ts;
-	const char *ident = make_hash_ident(needle->info);
-
-	for (ts = haystack[hash_djb2a(ident)]; ts; ts = ts->next) {
-		if (strcmp(ts->info->ident, ident) == 0)
-			return ts;
-	}
-
-	return NULL;	/* no success */
+	sample_add(ts_new, time_now);
 }
 
 /*
@@ -1127,12 +1198,12 @@ static void timer_stat_sort_freq_add(
 {
 	while (*sorted) {
 		if (opt_flags & OPT_CUMULATIVE) {
-			if ((*sorted)->count < new->count) {
+			if ((*sorted)->info->total_events < new->info->total_events) {
 				new->sorted_freq_next = *(sorted);
 				break;
 			}
 		} else {
-			if ((*sorted)->delta < new->delta) {
+			if ((*sorted)->info->delta_events < new->info->delta_events) {
 				new->sorted_freq_next = *(sorted);
 				break;
 			}
@@ -1164,54 +1235,31 @@ static void es_printf(const char *fmt, ...)
 }
 
 /*
- *  timer_stat_diff()
- *	find difference in event count between to hash table samples of timer
- *	stats.  We are interested in just current and new timers, not ones that
- *	silently die
+ *  timer_stat_dump()
  */
-static OPTIMIZE3 void timer_stat_diff(
+static OPTIMIZE3 void timer_stat_dump(
 	const double duration,		/* time between each sample */
 	const double time_delta,	/* how long been running sofar */
 	const int32_t n_lines,		/* number of lines to output */
 	const double whence,		/* nth sample */
-	timer_stat_t *timer_stats_old[],/* old timer stats samples */
-	timer_stat_t *timer_stats_new[])/* new timer stats samples */
+	timer_stat_t *timer_stats[])	/* timer stats samples */
 {
-	int i;
+	size_t i;
 	timer_stat_t *sorted = NULL;
 
 	for (i = 0; i < TABLE_SIZE; i++) {
 		timer_stat_t *ts;
 
-		for (ts = timer_stats_new[i]; ts; ts = ts->next) {
-			ts->info->last_used = whence;
-			timer_stat_t *found =
-				timer_stat_find(timer_stats_old, ts);
-			if (found) {
-				found->info->last_used = whence;
-				ts->delta = ts->count - found->count;
-				ts->time_delta = ts->time - found->time;
-				if (ts->delta >= opt_threshold) {
-					timer_stat_sort_freq_add(&sorted, ts);
-					sample_add(ts, whence);
-					found->info->total += ts->delta;
-					found->info->time_total += ts->time_delta;
-				}
-			} else {
-				ts->delta = ts->count;
-				ts->time_delta = duration;
-				if (ts->delta >= opt_threshold) {
-					timer_stat_sort_freq_add(&sorted, ts);
-					sample_add(ts, whence);
-				}
-			}
-		}
+		for (ts = timer_stats[i]; ts; ts = ts->next)
+			timer_stat_sort_freq_add(&sorted, ts);
 	}
 
 	if (!(opt_flags & OPT_QUIET)) {
 		uint64_t total = 0UL, kt_total = 0UL;
 		int32_t j = 0;
-		int sz, ta_size, if_size, cb_size;
+		const int pid_size = pid_max_digits();
+		int sz, ta_size, fn_size = 0;
+		int min_width;
 
 		eventstat_winsize();
 		if (resized && curses_init) {
@@ -1220,58 +1268,89 @@ static OPTIMIZE3 void timer_stat_diff(
 			resized = false;
 		}
 
-		sz = (cols - 80) / 6;
-		if (sz < 0)
-			sz = 0;
+		/* Minimum width w/o task or func info */
+		min_width = EVENTS_WIDTH + 1 + \
+			    1 + \
+			    pid_size + 1;
+		if (!(opt_flags & OPT_BRIEF)) {
+			if (opt_flags & OPT_TIMER_ID)
+				min_width += TIMER_ID_WIDTH + 1;
+			fn_size = FUNC_WIDTH;
+		}
 
-		ta_size = TASK_WIDTH + sz;
-		if_size = FUNC_WIDTH + (3 * sz);
-		cb_size = cols - (8 + 1 + 5 + 1 + ta_size + 1 + if_size + 2);
-		if (cb_size < 0)
-			cb_size = 20;
+		sz = cols - min_width;
+		sz = (sz < 0) ? 0 : sz;
 
-		es_printf("%8s %-5s %-*.*s",
+		if (fn_size) {
+			fn_size += (sz >> 1);
+			if (fn_size > FUNC_WIDTH_MAX)
+				fn_size = FUNC_WIDTH_MAX;
+
+			min_width += fn_size;
+			sz = cols - min_width;
+			sz = (sz < 0) ? 0 : sz;
+		}
+		ta_size = sz;
+		if (ta_size < TASK_WIDTH)
+			ta_size = TASK_WIDTH;
+		
+		es_printf("%*.*s %-*.*s %-*.*s",
+			EVENTS_WIDTH, EVENTS_WIDTH,
 			(opt_flags & OPT_CUMULATIVE) ?
-				"Events" : "Event/s", "PID",
-				ta_size, ta_size, "Task");
-		if (!(opt_flags & OPT_BRIEF))
-			es_printf(" %-*.*s %-*.*s\n",
-				if_size, if_size, "Init Function",
-				cb_size, cb_size, "Callback");
-		else
+				"Events" : "Event/s",
+			pid_size, pid_size, "PID",
+			ta_size, ta_size, "Task");
+		if (!(opt_flags & OPT_BRIEF)) {
+			if (opt_flags & OPT_TIMER_ID) {
+				es_printf(" %-16.16s", "Timer ID");
+			}
+			es_printf("%-*.*s\n", fn_size, fn_size,
+				" Init Function");
+		} else {
 			es_printf("\n");
+		}
 
 		while (sorted) {
 			if (((n_lines == -1) || (j < n_lines)) &&
-			    (sorted->delta != 0)) {
+			    (sorted->info->delta_events != 0)) {
 				char *task = (opt_flags & OPT_CMD) ?
-					sorted->info->cmdline : sorted->info->task_mangled;
+					sorted->info->cmdline :
+					sorted->info->task_mangled;
 				if (!*task)
 					task = sorted->info->task_mangled;
 
 				j++;
 				if (opt_flags & OPT_CUMULATIVE)
-					es_printf("%8" PRIu64 " ",
-						sorted->count);
+					es_printf("%*" PRIu64 " ",
+						EVENTS_WIDTH,
+						sorted->info->total_events);
 				else
-					es_printf("%8.2f ", (double)sorted->delta / duration);
+					es_printf("%*.2f ",
+						EVENTS_WIDTH,
+						(double)sorted->info->delta_events / duration);
 
 				if (opt_flags & OPT_BRIEF) {
 
-					es_printf("%5d %s\n",
-						sorted->info->pid, task);
+					es_printf("%*d %s\n",
+						pid_size, sorted->info->pid,
+						task);
 				} else {
-					es_printf("%5d %-*.*s %-*.*s %-*.*s\n",
-						sorted->info->pid,
-						ta_size, ta_size, task,
-						if_size, if_size, sorted->info->func,
-						cb_size, cb_size, sorted->info->callback);
+					es_printf("%*d %-*.*s",
+						pid_size, sorted->info->pid,
+						ta_size, ta_size, task);
+					if (opt_flags & OPT_TIMER_ID) {
+						es_printf(" %16" PRIx64,
+							sorted->info->timer);
+					}
+					es_printf(" %-*.*s\n",
+						fn_size - 1, fn_size - 1,
+						sorted->info->func);
 				}
 			}
-			total += sorted->delta;
+			total += sorted->info->delta_events;
 			if (sorted->info->kernel_thread)
-				kt_total += sorted->delta;
-
+				kt_total += sorted->info->delta_events;
+			sorted->info->delta_events = 0;
 			sorted = sorted->sorted_freq_next;
 		}
 		eventstat_move(LINES - 1, 0);
@@ -1305,7 +1384,7 @@ static OPTIMIZE3 void timer_stat_diff(
  *	only free'd at exit time.  This way we can parse
  *	the data a little faster.
  */
-static char *read_events(void)
+static char *read_events(const double time_end)
 {
 	int fd;
 	static size_t get_events_size;
@@ -1314,25 +1393,43 @@ static char *read_events(void)
 	if (get_events_buf == NULL) {
 		if ((get_events_buf = malloc(EVENT_BUF_SIZE << 1)) == NULL)
 			err_abort("Cannot read %s, out of memory\n",
-				proc_timer_stats);
+				sys_tracing_pipe);
 
 		get_events_size = (EVENT_BUF_SIZE << 1);
 	}
 
-	if ((fd = open(proc_timer_stats, O_RDONLY)) < 0)
-		err_abort("Cannot open %s\n", proc_timer_stats);
+	if ((fd = open(sys_tracing_pipe, O_RDONLY)) < 0)
+		err_abort("Cannot open %s\n", sys_tracing_pipe);
 
 	size = 0;
-	for (;;) {
+	while (!stop_eventstat) {
 		ssize_t ret;
-		char buffer[EVENT_BUF_SIZE];
+		int rc;
+		static char buffer[EVENT_BUF_SIZE];
+		const double duration = time_end - gettime_to_double();
+		struct timeval tv;
+		fd_set rfds;
 
+		if (duration < 0.0)
+			break;
+
+		tv = double_to_timeval(duration);
+		FD_ZERO(&rfds);
+		FD_SET(fd, &rfds);
+
+		errno = 0;
+		rc = select(fd + 1, &rfds, NULL, NULL, &tv);
+		if (rc <= 0)
+			break;
+		if (!FD_ISSET(fd, &rfds))
+			continue;
 		ret = read(fd, buffer, sizeof(buffer));
 		if (ret == 0)
-			break;
+			continue;
 		if (ret < 0) {
-			if ((errno == EINTR) ||
-			    (errno != EAGAIN)) {
+			if (!stop_eventstat &&
+			    ((errno == EINTR) ||
+			     (errno != EAGAIN))) {
 				continue;
 			}
 			break;
@@ -1346,7 +1443,7 @@ static char *read_events(void)
 			if (!tmpptr) {
 				(void)close(fd);
 				err_abort("Cannot read %s, out of memory\n",
-					proc_timer_stats);
+					sys_tracing_pipe);
 			}
 			get_events_buf = tmpptr;
 		}
@@ -1361,25 +1458,26 @@ static char *read_events(void)
 
 /*
  *  get_events()
- *	scan /proc/timer_stats and populate a timer stat hash table with
- *	unique events
+ *	parse /sys/kernel/debug/tracing/trace_pipe and populate
+ *	a timer stat hash table with unique events
  */
 static void get_events(
 	timer_stat_t *timer_stats[],
-	const double time_now)
+	const double time_now,
+	const double duration)
 {
-	char *tmpptr;
 	const size_t app_name_len = strlen(app_name);
+	const double time_end = time_now + duration - 0.05;
+	char *tmpptr = read_events(time_end);
 
-	if ((tmpptr = read_events()) == NULL)
+	if (!tmpptr)
 		return;
 
 	while (*tmpptr) {
-		char *ptr = tmpptr, *eol = tmpptr;
+		char *ptr, *eol = tmpptr;
 		char task[64];
 		char task_mangled[64];
 		char func[64];
-		char callback[64];
 		char *cmdline;
 		int mask;
 		timer_info_t info;
@@ -1393,36 +1491,43 @@ static void get_events(
 			}
 			eol++;
 		}
+		if (strstr(tmpptr, "hrtimer_start")) {
+			memset(&info, 0, sizeof(info));
+			memset(task, 0, sizeof(task));
+			memset(func, 0, sizeof(func));
 
-		/* Is this the last line we want to parse? */
-		if (strstr(tmpptr, "total events") != NULL)
-			break;
+			/*
+			 *  Parse something like the following:
+			 *  gnome-shell-3515  [003] d.h. 101499.108349: hrtimer_start: hrtimer=ffff99979e2d4600 function=tick_sched_timer expires=101497144000000 softexpires=101497144000000
+			 */
+			if (sscanf(tmpptr, "%s %*s %*s %*f: hrtimer_start: hrtimer=%" PRIx64 " function=%s", task, &info.timer, func) != 3)
+				goto next;
+		} else {
+			goto next;
+		}
 
-		/* Looking for a format: "count[D], pid, task, func (timer)" */
-		while (*ptr && *ptr != ',')
+		/*
+		 * task name in form: gnome-shell-3515, scan to end of
+		 * string, then scan back to find start of PID
+		 */
+		ptr = task;
+		while (*ptr)
 			ptr++;
-		if (*ptr != ',')
-			goto next;
-		if (ptr > tmpptr && *(ptr - 1) == 'D')
-			goto next;	/* Deferred event, skip */
-
-		/* Now we're ready to fetch info fields */
-		memset(&info, 0, sizeof(info));
+		ptr--;
+		while (ptr >= task && (*ptr >= '0' && *ptr <= '9'))
+			ptr--;
 
+		*ptr = '\0';
 		ptr++;
-		if (sscanf(tmpptr, "%21" SCNu64, &info.total) != 1)
+
+		if (sscanf(ptr, "%10d\n", &info.pid) != 1)
 			goto next;
-		memset(task, 0, sizeof(task));
-		memset(func, 0, sizeof(func));
-		memset(callback, 0, sizeof(callback));
-		info.pid = -1;
-		if (sscanf(ptr, "%10d %63s %63s (%63[^)])",
-		    &info.pid, task, func, callback) != 4)
+		if (info.pid == 0)
 			goto next;
 
 		/* Processes without a command line are kernel threads */
 		cmdline = get_pid_cmdline(info.pid);
-		info.kernel_thread = pid_a_kernel_thread(cmdline, task, info.pid);
+		info.kernel_thread = pid_a_kernel_thread(task, info.pid);
 
 		/* Swapper is special, like all corner cases */
 		if (strncmp(task, "swapper", 6) == 0)
@@ -1437,27 +1542,21 @@ static void get_events(
 
 			strcpy(tmp, task);
 			tmp[13] = '\0';
-			snprintf(task_mangled, sizeof(task_mangled), "[%s]", tmp);
+			snprintf(task_mangled, sizeof(task_mangled),
+				"[%s]", tmp);
 		} else {
 			strcpy(task_mangled, task);
 		}
 
-		if (strcmp(task, "insmod") == 0)
-			strncpy(task, "[kern mod]", 13);
-		if (strcmp(task, "modprobe") == 0)
-			strncpy(task, "[kern mod]", 13);
-
-		if (strncmp(func, "tick_nohz_", 10) &&
-		    strncmp(func, "tick_setup_sched_timer", 20) &&
-		    strncmp(task, app_name, app_name_len)) {
+		if (strncmp(task, app_name, app_name_len)) {
 			info.task = task;
 			info.cmdline = cmdline ? cmdline : task_mangled;
 			info.task_mangled = task_mangled;
 			info.func = func;
-			info.callback = callback;
-			info.ident = tmpptr;
 			info.time_total = 0.0;
-			timer_stat_add(timer_stats, time_now, &info);
+			info.total_events = 1;
+			info.ident = make_hash_ident(&info);
+			timer_stat_add(timer_stats, &info, time_now, duration);
 		}
 free_next:
 		free(cmdline);
@@ -1507,7 +1606,7 @@ static void handle_sigwinch(int sig)
 
 int main(int argc, char **argv)
 {
-	timer_stat_t **timer_stats_old, **timer_stats_new, **tmp;
+	timer_stat_t **timer_stats;
 	double duration_secs = 1.0, time_start, time_now;
 	int64_t count = 1, t = 1;
 	int32_t n_lines = -1;
@@ -1517,7 +1616,7 @@ int main(int argc, char **argv)
 	int i;
 
 	for (;;) {
-		int c = getopt(argc, argv, "bcCdksSlhn:qr:t:Tuw");
+		int c = getopt(argc, argv, "bcCdksSlhin:qr:t:Tuw");
 		if (c == -1)
 			break;
 		switch (c) {
@@ -1536,6 +1635,9 @@ int main(int argc, char **argv)
 		case 'h':
 			show_usage();
 			eventstat_exit(EXIT_SUCCESS);
+		case 'i':
+			opt_flags |= OPT_TIMER_ID;
+			break;
 		case 'n':
 			errno = 0;
 			n_lines = (int32_t)strtol(optarg, NULL, 10);
@@ -1545,6 +1647,9 @@ int main(int argc, char **argv)
 			if (n_lines < 1)
 				err_abort("-n option must be greater than 0\n");
 			break;
+		case 'S':
+			opt_flags |= OPT_RESULT_STATS;
+			break;
 		case 't':
 			opt_threshold = strtoull(optarg, NULL, 10);
 			if (opt_threshold < 1)
@@ -1562,9 +1667,6 @@ int main(int argc, char **argv)
 		case 's':
 			opt_flags |= OPT_CMD_SHORT;
 			break;
-		case 'S':
-			opt_flags |= OPT_RESULT_STATS;
-			break;
 		case 'l':
 			opt_flags |= OPT_CMD_LONG;
 			break;
@@ -1605,8 +1707,8 @@ int main(int argc, char **argv)
 	opt_threshold *= duration_secs;
 
 	if (geteuid() != 0)
-		err_abort("%s requires root privileges to write to %s\n",
-			app_name, proc_timer_stats);
+		err_abort("%s requires root privileges to gather "
+			"trace event data\n", app_name);
 
 	sane_procs = sane_proc_pid_info();
 	if (!sane_procs)
@@ -1623,16 +1725,14 @@ int main(int argc, char **argv)
 				errno, strerror(errno));
 	}
 
-	if ((timer_stats_old = calloc(TABLE_SIZE, sizeof(timer_stat_t*))) == NULL)
-		err_abort("Cannot allocate old timer stats table\n");
-	if ((timer_stats_new = calloc(TABLE_SIZE, sizeof(timer_stat_t*))) == NULL)
-		err_abort("Cannot allocate old timer stats table\n");
+	if ((timer_stats = calloc(TABLE_SIZE, sizeof(timer_stat_t*))) == NULL)
+		err_abort("Cannot allocate timer stats table\n");
 
 	/* Should really catch signals and set back to zero before we die */
-	set_timer_stat("1\n", true);
-	time_now = time_start = gettime_to_double();
+	set_tracing_enable("1\n", true);
+	set_tracing_event();
 
-	get_events(timer_stats_old, time_now);
+	time_now = time_start = gettime_to_double();
 
 	if (opt_flags & OPT_TOP) {
 		struct sigaction sa;
@@ -1652,9 +1752,7 @@ int main(int argc, char **argv)
 	}
 
 	while (!stop_eventstat && (forever || count--)) {
-		struct timeval tv;
 		double secs, duration, time_delta;
-		int ret;
 
 		/* Timeout to wait for in the future for this sample */
 		secs = time_start + ((double)t * duration_secs) - time_now;
@@ -1669,12 +1767,15 @@ int main(int argc, char **argv)
 			if (!redo)
 				t++;
 		}
+
 		redo = false;
-		tv = double_to_timeval(secs);
 
 		if (curses_init) {
 			fd_set rfds;
-			int ch;
+			int ch, ret;
+			struct timeval tv;
+
+			memset(&tv, 0, sizeof(tv));
 
 			FD_ZERO(&rfds);
 			FD_SET(fileno(stdin), &rfds);
@@ -1685,48 +1786,38 @@ int main(int argc, char **argv)
 				break;
 			if (ret > 0)
 				redo = true;
-		} else {
-			ret = select(0, NULL, NULL, NULL, &tv);
-		}
+			if (ret < 0) {
+				if (errno != EINTR) {
+					eventstat_endwin();
 
-		if (ret < 0) {
-			if (errno != EINTR) {
-				eventstat_endwin();
-
-				fprintf(stderr, "select() failed: "
-					"errno=%d (%s)\n",
-					errno, strerror(errno));
-				goto abort;
+					fprintf(stderr, "select() failed: "
+						"errno=%d (%s)\n",
+						errno, strerror(errno));
+					goto abort;
+				}
+				redo = true;
 			}
-			redo = true;
 		}
 
+		get_events(timer_stats, time_now, secs);
+
 		duration = gettime_to_double() - time_now;
 		duration = floor((duration * 1000.0) + 0.5) / 1000.0;
 		time_now = gettime_to_double();
 		time_delta = time_now - time_start;
 
-		get_events(timer_stats_new, time_now);
 		eventstat_clear();
-		timer_stat_diff(duration, time_delta, n_lines, time_now,
-			timer_stats_old, timer_stats_new);
+		timer_stat_dump(duration, time_delta, n_lines, time_now, timer_stats);
 		eventstat_refresh();
-		timer_stat_free_contents(timer_stats_old);
-
-		tmp             = timer_stats_old;
-		timer_stats_old = timer_stats_new;
-		timer_stats_new = tmp;
-
+		timer_stat_free_contents(timer_stats);
 		timer_info_purge_old(time_now);
 	}
 	eventstat_endwin();
 abort:
 	samples_dump(csv_results);
 
-	timer_stat_free_contents(timer_stats_old);
-	timer_stat_free_contents(timer_stats_new);
-	free(timer_stats_old);
-	free(timer_stats_new);
+	timer_stat_free_contents(timer_stats);
+	free(timer_stats);
 	samples_free();
 	timer_info_list_free();
 	timer_stat_free_list_free();
